
{*******************************************************}
{                                                       }
{       Borland Delphi Demo Program Library             }
{                                                       }
{       Copyright (c) 1995,99 Inprise Corporation       }
{                                                       }
{       modified for Delphi 4, 5 WDosX DOS-Extender     }
{       Copyright (c) 2000 by Immo Wache                }
{       e-mail: immo.wache@t-online.de                  }
{                                                       }
{*******************************************************}

{
  Version history:
    IWA 11/20/00 Version 1.1
}

unit WDosExeImage;

interface

uses
  WDosTypInfo, WDosClasses, WDosSysUtils, WDosWindows, WDosRXTypes;

const
  MF_POPUP = $10;
  MF_END = $80;

type

{ Exceptions }

  EExeError = class (Exception)
  end;

{ Forward Declarations }

  TResourceItem = class;
  TResourceClass = class of TResourceItem;
  TResourceList = class;

{ TExeImage }

  TExeImage = class (TComponent)
  private
    FCursorResources: TResourceItem;
    FDosHeader: PIMAGE_DOS_HEADER;
    FIconResources: TResourceItem;
    FNTHeader: PIMAGE_NT_HEADERS;
    FRCDataResources: TResourceItem;
    FResourceBase: LongInt;
    FResourceList: TResourceList;
    FResourceRVA: LongInt;
    FStream: TStream;
    function GetRCDataResources: TResourceItem;
    function GetResourceList: TResourceList;
    function GetSectionHdr(const SectionName: string; var Header:  
        PIMAGE_SECTION_HEADER): Boolean;
  public
    constructor CreateImage(AOwner: TComponent; AStream: TStream);
    destructor Destroy; override;
    function GetResources(AResType: TResourceType): TResourceItem;
    property RCDataResources: TResourceItem read GetRCDataResources;
    property ResourceList: TResourceList read GetResourceList;
  end;

{ TResourceItem }

  TResourceItem = class (TComponent)
  private
    FDataEntry: PIMAGE_RESOURCE_DATA_ENTRY;
    FDirEntry: PIMAGE_RESOURCE_DIRECTORY_ENTRY;
    FFirstChildDirEntry: PIMAGE_RESOURCE_DIRECTORY_ENTRY;
    FList: TResourceList;
    FRawData: Pointer;
    FRawDataSize: Integer;
    function DataEntry: PIMAGE_RESOURCE_DATA_ENTRY;
    function FExeImage: TExeImage;
    function FirstChildDirEntry: PIMAGE_RESOURCE_DIRECTORY_ENTRY;
    function GetRawData: Pointer;
    function GetResourceItem(Index: Integer): TResourceItem;
    function GetResourceType: TResourceType;
  protected
    function GetName: string; virtual;
    function GetResourceList: TResourceList; virtual;
  public
    constructor CreateItem(AOwner: TComponent; ADirEntry:  
        PIMAGE_RESOURCE_DIRECTORY_ENTRY);
    destructor Destroy; override;
    procedure FreeRawData;
    function IsList: Boolean; virtual;
    function Offset: Integer;
    function RawDataOfs: Integer;
    function ResTypeStr: string;
    procedure SaveToFile(const FileName: string);
    procedure SaveToStream(Stream: TStream); virtual;
    function Size: Integer;
    property Items[Index: Integer]: TResourceItem read GetResourceItem; default;
    property List: TResourceList read GetResourceList;
    property Name: string read GetName;
    property RawData: Pointer read GetRawData;
    property ResType: TResourceType read GetResourceType;
  end;

{ TIconResource }

  TIconResource = class (TResourceItem)
  protected
    function GetResourceList: TResourceList; override;
  public
    function IsList: Boolean; override;
  end;

{ TIconResEntry }

  TIconResEntry = class (TResourceItem)
  protected
    FResInfo: PIconResInfo;
    function GetName: string; override;
  end;

{ TCursorResource }

  TCursorResource = class (TIconResource)
  protected
    function GetResourceList: TResourceList; override;
  end;

{ TCursorResEntry }

  TCursorResEntry = class (TIconResEntry)
  protected
    FResInfo: PCursorResInfo;
    function GetName: string; override;
  end;

{ TBitmapResource }

  TBitMapResource = class (TResourceItem)
  end;

{ TStringResource }

  TStringResource = class (TResourceItem)
  protected
    procedure AssignTo(Dest: TPersistent); override;
  end;

{ TMenuResource }

  TMenuResource = class (TResourceItem)
  private
    FNestLevel: Integer;
    FNestStr: string;
    procedure SetNestLevel(Value: Integer);
  protected
    procedure AssignTo(Dest: TPersistent); override;
    property NestLevel: Integer read FNestLevel write SetNestLevel;
    property NestStr: string read FNestStr;
  end;

{ TResourceList }

  TResourceList = class (TComponent)
  private
    function GetResourceType: TResourceType;
  protected
    FExeImage: TExeImage;
    FList: TList;
    FResDir: PIMAGE_RESOURCE_DIRECTORY;
    FResDirOfs: Integer;
    FResType: Integer;
    function GetResourceItem(Index: Integer): TResourceItem;
    function List: TList; virtual;
  public
    constructor CreateList(AOwner: TComponent; ResDirOfs: Longint; AExeImage:  
        TExeImage);
    destructor Destroy; override;
    function Count: Integer;
    function FindResourceItem(ResourceName: string): TResourceItem;
    property Items[Index: Integer]: TResourceItem read GetResourceItem; default;
    property ResType: TResourceType read GetResourceType;
  end;

{ TIconResourceList }

  TIconResourceList = class (TResourceList)
  protected
    function List: TList; override;
  end;

{ TCursorResourceList }

  TCursorResourceList = class (TResourceList)
  protected
    function List: TList; override;
  end;

implementation

function lstrlenW(lpString: PWideChar): Integer;
begin
  Result :=0;
  while lpString[Result] <>#0 do
    Inc(Result);
end;

{ This function maps a resource type to the associated resource class }

function GetResourceClass(ResType: Integer): TResourceClass;
const
  TResourceClasses: array[TResourceType] of TResourceClass = (
    TResourceItem,      { rtUnknown0 }
    TCursorResEntry,    { rtCursorEntry }
    TBitmapResource,    { rtBitmap }
    TIconResEntry,      { rtIconEntry }
    TMenuResource,      { rtMenu }
    TResourceItem,      { rtDialog }
    TStringResource,    { rtString }
    TResourceItem,      { rtFontDir }
    TResourceItem,      { rtFont }
    TResourceItem,      { rtAccelerators }
    TResourceItem,      { rtRCData }
    TResourceItem,      { rtMessageTable }
    TCursorResource,    { rtGroupCursor }
    TResourceItem,      { rtUnknown13 }
    TIconResource,      { rtIcon }
    TResourceItem,      { rtUnknown15 }
    TResourceItem);     { rtVersion }
begin
  if (ResType >= Integer(Low(TResourceType))) and
    (ResType <= Integer(High(TResourceType))) then
    Result := TResourceClasses[TResourceType(ResType)] else
    Result := TResourceItem;
end;

{ Utility Functions }

function Min(A, B: Integer): Integer;
begin
  if A < B then Result := A
  else Result := B;
end;

{ This function checks if an offset is a string name, or a directory }
{Assumes: IMAGE_RESOURCE_NAME_IS_STRING = IMAGE_RESOURCE_DATA_IS_DIRECTORY}

function HighBitSet(L: Longint): Boolean;
begin
  Result := (L and IMAGE_RESOURCE_DATA_IS_DIRECTORY) <> 0;
end;

function StripHighBit(L: Longint): Longint;
begin
  Result := L and IMAGE_OFFSET_STRIP_HIGH;
end;

function StripHighPtr(L: Longint): Pointer;
begin
  Result := Pointer(L and IMAGE_OFFSET_STRIP_HIGH);
end;

{ This function converts a pointer to a wide char string into a pascal string }

function WideCharToStr(WStr: PWChar; Len: Integer): string;
var
  I: Integer;
begin
  SetLength(Result, Len);
  for I :=1 to Len do
    Result[I] :=Chr(Ord(WStr[I -1]));
end;

{ Exceptions }

procedure ExeError(const ErrMsg: string);
begin
  raise EExeError.Create(ErrMsg);
end;

{
********************************** TExeImage ***********************************
}
constructor TExeImage.CreateImage(AOwner: TComponent; AStream: TStream);
begin
  inherited Create(AOwner);
  FStream :=AStream;
  
  New(FDosHeader);
  FStream.ReadBuffer(FDosHeader^, SizeOf(IMAGE_DOS_HEADER));
  if not FDosHeader.e_magic = IMAGE_DOS_SIGNATURE then
    ExeError('unrecognized file format');
  
  New(FNTHeader);
  FStream.Seek(FDosHeader.e_lfanew, soFromBeginning);
  FStream.ReadBuffer(FNTHeader^, SizeOf(IMAGE_NT_HEADERS));
  if FNTHeader.Signature <> IMAGE_NT_SIGNATURE then
    ExeError('Not a PE (WIN32 Executable) file');
end;

destructor TExeImage.Destroy;
begin
  Dispose(FNTHeader);
  Dispose(FDosHeader);
  FStream.Free;
  inherited Destroy;
end;

function TExeImage.GetResources(AResType: TResourceType): TResourceItem;
var
  I: Integer;
begin
  for I :=0 to ResourceList.Count -1 do
    with ResourceList[I] do
    begin
      Result :=ResourceList[I];
      if IsList and (List.ResType =AResType) then
        Exit;
    end;
  Result :=nil;
end;

function TExeImage.GetSectionHdr(const SectionName: string; var Header:  
    PIMAGE_SECTION_HEADER): Boolean;
var
  I: Integer;
begin
  FStream.Seek(FDosHeader.e_lfanew +SizeOf(IMAGE_NT_HEADERS),
    soFromBeginning);
  FStream.ReadBuffer(Header^, SizeOf(IMAGE_SECTION_HEADER));
  Result := True;
  for I := 0 to FNTHeader.FileHeader.NumberOfSections - 1 do
  begin
    if StrLIComp(Header.Name, PChar(SectionName), IMAGE_SIZEOF_SHORT_NAME) = 0 then Exit;
    FStream.ReadBuffer(Header^, SizeOf(IMAGE_SECTION_HEADER));
  end;
  Result := False;
end;

function TExeImage.GetRCDataResources: TResourceItem;
begin
  if not Assigned(FRCDataResources) then
    FRCDataResources :=GetResources(rtRCData);
  Result :=FRCDataResources;
end;

function TExeImage.GetResourceList: TResourceList;
var
  ResSectHdr: PIMAGE_SECTION_HEADER;
begin
  if not Assigned(FResourceList) then
  begin
    New(ResSectHdr);
    try
      if GetSectionHdr('.rsrc', ResSectHdr) then
      begin
        FResourceBase := ResSectHdr.PointerToRawData;
        FResourceRVA := ResSectHdr.VirtualAddress;
        FResourceList := TResourceList.CreateList(Self, FResourceBase, Self);
      end
      else
        ExeError('No resources in this file.');
    finally
      Dispose(ResSectHdr);
    end;
  end;
  Result := FResourceList;
end;

{
******************************** TResourceItem *********************************
}
constructor TResourceItem.CreateItem(AOwner: TComponent; ADirEntry:  
    PIMAGE_RESOURCE_DIRECTORY_ENTRY);
begin
  inherited Create(AOwner);
  New(FDirEntry);
  FDirEntry^ :=ADirEntry^;
end;

destructor TResourceItem.Destroy;
begin
  FreeRawData;
  FreeMem(FDataEntry);
  FreeMem(FFirstChildDirEntry);
  Dispose(FDirEntry);
  inherited Destroy;
end;

function TResourceItem.DataEntry: PIMAGE_RESOURCE_DATA_ENTRY;
begin
  if not Assigned(FDataEntry) then
  begin
    New(FDataEntry);
    with FExeImage.FStream do
    begin
      Seek(FirstChildDirEntry.OffsetToData + Cardinal(FExeImage.FResourceBase),
          soFromBeginning);
      ReadBuffer(FDataEntry^, SizeOf(IMAGE_RESOURCE_DATA_ENTRY));
    end;
  end;
  Result :=FDataEntry;
end;

function TResourceItem.FExeImage: TExeImage;
begin
  Result := (Owner as TResourceList).FExeImage;
end;

function TResourceItem.FirstChildDirEntry: PIMAGE_RESOURCE_DIRECTORY_ENTRY;
begin
  if not Assigned(FFirstChildDirEntry) then
  begin
    New(FFirstChildDirEntry);
    with FExeImage.FStream do
    begin
      Seek(StripHighBit(LongInt(FDirEntry.OffsetToData)) +FExeImage.FResourceBase
          +SizeOf(IMAGE_RESOURCE_DIRECTORY), soFromBeginning);
      ReadBuffer(FFirstChildDirEntry^, SizeOf(IMAGE_RESOURCE_DIRECTORY_ENTRY));
    end;
  end;
  Result :=FFirstChildDirEntry;
end;

procedure TResourceItem.FreeRawData;
begin
  FreeMem(FRawData, FRawDataSize);
  FRawData :=nil;
  FRawDataSize :=0;
end;

function TResourceItem.IsList: Boolean;
begin
  Result := HighBitSet(LongInt(FirstChildDirEntry.OffsetToData));
end;

function TResourceItem.Offset: Integer;
begin
  if IsList then
    Result := StripHighBit(LongInt(FDirEntry.OffsetToData))
  else
    Result := DataEntry.OffsetToData;
end;

function TResourceItem.RawDataOfs: Integer;
begin
  with FExeImage do
    Result :=FResourceBase -FResourceRVA +LongInt(DataEntry.OffsetToData);
end;

function TResourceItem.ResTypeStr: string;
begin
  Result := Copy(GetEnumName(TypeInfo(TResourceType), Ord(ResType)), 3, 20);
end;

procedure TResourceItem.SaveToFile(const FileName: string);
var
  FS: TFileStream;
begin
  FS := TFileStream.Create(FileName, fmCreate);
  try
    Self.SaveToStream(FS);
  finally
    FS.Free;
  end;
end;

procedure TResourceItem.SaveToStream(Stream: TStream);
begin
  Stream.Write(RawData^, Size);
end;

function TResourceItem.Size: Integer;
begin
  if IsList then
    Result := 0
  else
    Result := DataEntry.Size;
end;

function TResourceItem.GetName: string;
var
  Length: Word;
  NameString: PWChar;
begin
  { Check for Level1 entries, these are resource types. }
  if (Owner.Owner = FExeImage) and not HighBitSet(LongInt(FDirEntry.Name)) and
    (FDirEntry.Name <= 16) then
  begin
    Result := Copy(GetEnumName(TypeInfo(TResourceType), FDirEntry.Name), 3, 20);
    Exit;
  end;
  if HighBitSet(LongInt(FDirEntry.Name)) then
  begin
    with FExeImage.FStream do
    begin
      Seek(StripHighBit(LongInt(FDirEntry.Name)) +FExeImage.FResourceBase,
          soFromBeginning);
      ReadBuffer(Length, SizeOf(Length));
      GetMem(NameString, Length *SizeOf(WChar));
      try
        ReadBuffer(NameString^, Length *SizeOf(WChar));
        Result := WideCharToStr(NameString, Length);
      finally
        FreeMem(NameString, Length *SizeOf(WChar));
      end;
    end;
  end
  else Result := Format('%d', [FDirEntry.Name]);
end;

function TResourceItem.GetRawData: Pointer;
begin
  if not Assigned(FRawData) and (Size >0) then
  begin
    FRawDataSize :=Size;
    GetMem(FRawData, FRawDataSize);
    with FExeImage.FStream do
    begin
      Seek(RawDataOfs, soFromBeginning);
      ReadBuffer(FRawData^, FRawDataSize);
    end;
  end;
  Result :=FRawData;
end;

function TResourceItem.GetResourceItem(Index: Integer): TResourceItem;
begin
  Result := List[Index];
end;

function TResourceItem.GetResourceList: TResourceList;
begin
  if not IsList then ExeError('ResourceItem is not a list');
  if not Assigned(FList) then
    FList := TResourceList.CreateList(Self,
        StripHighBit(LongInt(FDirEntry.OffsetToData))
        +FExeImage.FResourceBase, FExeImage);
  Result := FList;
end;

function TResourceItem.GetResourceType: TResourceType;
begin
  Result := TResourceType((Owner as TResourceList).FResType);
end;

{
******************************** TIconResource *********************************
}
function TIconResource.GetResourceList: TResourceList;
begin
  if not Assigned(FList) then
    FList := TIconResourceList.CreateList(Owner, RawDataOfs, FExeImage);
  Result := FList;
end;

function TIconResource.IsList: Boolean;
begin
  Result := True;
end;

{
******************************** TIconResEntry *********************************
}
function TIconResEntry.GetName: string;
begin
  if Assigned(FResInfo) then
    with FResInfo^ do
      Result := Format('%d X %d %d Colors', [bWidth, bHeight, bColorCount])
  else
    Result := inherited GetName;
end;

{
******************************* TCursorResource ********************************
}
function TCursorResource.GetResourceList: TResourceList;
begin
  if not Assigned(FList) then
    FList := TCursorResourceList.CreateList(Owner, RawDataOfs, FExeImage);
  Result := FList;
end;

{
******************************* TCursorResEntry ********************************
}
function TCursorResEntry.GetName: string;
begin
  if Assigned(FResInfo) then
    with FResInfo^ do
      Result := Format('%d X %d %d Bit(s)', [wWidth, wWidth, wBitCount])
  else
    Result := inherited GetName;
end;

{
******************************* TStringResource ********************************
}
procedure TStringResource.AssignTo(Dest: TPersistent);
var
  P: PWChar;
  ID: Integer;
  Cnt: Cardinal;
  Len: Word;
begin
  if (Dest is TStrings) then
    with TStrings(Dest) do
    begin
      BeginUpdate;
      try
        Clear;
        P := RawData;
        Cnt := 0;
        while Cnt < StringsPerBlock do
        begin
          Len := Word(P^);
          if Len > 0 then
          begin
            Inc(P);
            ID := ((FDirEntry.Name - 1) shl 4) + Cnt;
            Add(Format('%d,  "%s"', [ID, WideCharToStr(P, Len)]));
            Inc(P, Len);
          end;
          Inc(Cnt);
        end;
      finally
        EndUpdate;
      end;
    end
  else
    inherited AssignTo(Dest);
end;

{
******************************** TMenuResource *********************************
}
procedure TMenuResource.AssignTo(Dest: TPersistent);
var
  IsPopup: Boolean;
  Len: Word;
  MenuData: PWord;
  MenuEnd: PChar;
  MenuText: PWChar;
  MenuID: Word;
  MenuFlags: Word;
  S: string;
begin
  if (Dest is TStrings) then
    with TStrings(Dest) do
    begin
      BeginUpdate;
      try
        Clear;
        MenuData := RawData;
        MenuEnd := PChar(RawData) + Size;
        Inc(MenuData, 2);
        NestLevel := 0;
        while PChar(MenuData) < MenuEnd do
        begin
          MenuFlags := MenuData^;
          Inc(MenuData);
          IsPopup := (MenuFlags and MF_POPUP) = MF_POPUP;
          MenuID := 0;
          if not IsPopup then
          begin
            MenuID := MenuData^;
            Inc(MenuData);
          end;
          MenuText := PWChar(MenuData);
          Len := lstrlenW(MenuText);
          if Len = 0 then
            S := 'MENUITEM SEPARATOR'
          else
          begin
            S := WideCharToStr(MenuText, Len);
            if IsPopup then
              S := Format('POPUP "%s"', [S]) else
              S := Format('MENUITEM "%s",  %d', [S, MenuID]);
          end;
          Inc(MenuData, Len + 1);
          Add(NestStr + S);
          if (MenuFlags and MF_END) = MF_END then
          begin
            NestLevel := NestLevel - 1;
            Add(NestStr + 'ENDPOPUP');
          end;
          if IsPopup then
            NestLevel := NestLevel + 1;
        end;
      finally
        EndUpdate;
      end;
    end
  else
    inherited AssignTo(Dest);
end;

procedure TMenuResource.SetNestLevel(Value: Integer);
begin
  FNestLevel := Value;
  SetLength(FNestStr, Value * 2);
  FillChar(FNestStr[1], Value * 2, ' ');
end;

{
******************************** TResourceList *********************************
}
constructor TResourceList.CreateList(AOwner: TComponent; ResDirOfs: Longint; 
    AExeImage:  TExeImage);
begin
  inherited Create(AOwner);
  FExeImage := AExeImage;
  FResDirOfs :=ResDirOfs;
  New(FResDir);
  with FExeImage.FStream do
  begin
    Seek(FResDirOfs, soFromBeginning);
    ReadBuffer(FResDir^, SizeOf(IMAGE_RESOURCE_DIRECTORY));
  end;
  if AOwner <> AExeImage then
    if AOwner.Owner.Owner = AExeImage then
      FResType := TResourceItem(Owner).FDirEntry.Name
    else
      FResType := (AOwner.Owner.Owner as TResourceList).FResType;
end;

destructor TResourceList.Destroy;
begin
  Dispose(FResDir);
  inherited Destroy;
  FList.Free;
end;

function TResourceList.Count: Integer;
begin
  Result := List.Count;
end;

function TResourceList.FindResourceItem(ResourceName: string): TResourceItem;
var
  I: Integer;
begin
  ResourceName :=UpperCase(ResourceName);
  for I :=0 to Count -1 do
  begin
    Result :=Items[I];
    with Result do
      if (Name =ResourceName) and not IsList then Exit;
  end;
  Result :=nil;
end;

function TResourceList.List: TList;
var
  I: Integer;
  DirEntry: PIMAGE_RESOURCE_DIRECTORY_ENTRY;
  DirCnt: Integer;
  ResItem: TResourceItem;
begin
  if not Assigned(FList) then
  begin
    FList := TList.Create;
    New(DirEntry);
    try
      with FExeImage.FStream do
      begin
        Seek(FResDirOfs +SizeOf(IMAGE_RESOURCE_DIRECTORY), soFromBeginning);
        ReadBuffer(DirEntry^, SizeOf(IMAGE_RESOURCE_DIRECTORY_ENTRY));
      end;
      DirCnt := FResDir.NumberOfNamedEntries + FResDir.NumberOfIdEntries - 1;
      for I := 0 to DirCnt do
      begin
        { Handle Cursors and Icons specially }
        ResItem := GetResourceClass(FResType).CreateItem(Self, DirEntry);
        if Owner = FExeImage then
          if TResourceType(DirEntry.Name) in [rtCursorEntry, rtIconEntry] then
          begin
            if TResourceType(DirEntry.Name) = rtCursorEntry then
              FExeImage.FCursorResources := ResItem
            else
              FExeImage.FIconResources := ResItem;
            FExeImage.FStream.ReadBuffer(DirEntry^,
                SizeOf(IMAGE_RESOURCE_DIRECTORY_ENTRY));
            Continue;
          end;
        FList.Add(ResItem);
        FExeImage.FStream.ReadBuffer(DirEntry^,
            SizeOf(IMAGE_RESOURCE_DIRECTORY_ENTRY));
      end;
    finally
      Dispose(DirEntry);
    end;
  end;
  Result := FList;
end;

function TResourceList.GetResourceItem(Index: Integer): TResourceItem;
begin
  Result := List[Index];
end;

function TResourceList.GetResourceType: TResourceType;
begin
  Result := TResourceType(FResType);
end;

{
****************************** TIconResourceList *******************************
}
function TIconResourceList.List: TList;
var
  I,  J, Cnt: Integer;
  ResData: PIconResInfo;
  ResList: TResourceList;
  ResOrd: Cardinal;
  IconResource: TIconResEntry;
begin
  if not Assigned(FList) then
  begin
    FList := TList.Create;
    Cnt := PIconHeader(FResDir).wCount;
    PChar(ResData) := PChar(FResDir) + SizeOf(TIconHeader);
    ResList := FExeImage.FIconResources.List;
    for I := 0 to Cnt - 1 do
    begin
      ResOrd := ResData.wNameOrdinal;
      for J := 0 to ResList.Count - 1 do
      begin
        if ResOrd = ResList[J].FDirEntry.Name then
        begin
          IconResource := ResList[J] as TIconResEntry;
          IconResource.FResInfo := ResData;
          FList.Add(IconResource);
        end;
      end;
      Inc(ResData);
    end;
  end;
  Result := FList;
end;

{
***************************** TCursorResourceList ******************************
}
function TCursorResourceList.List: TList;
var
  I, J, Cnt: Integer;
  ResData: PCursorResInfo;
  ResList: TResourceList;
  ResOrd: Cardinal;
  CursorResource: TCursorResEntry;
begin
  if not Assigned(FList) then
  begin
    FList := TList.Create;
    Cnt := PIconHeader(FResDir).wCount;
    PChar(ResData) := PChar(FResDir) + SizeOf(TIconHeader);
    ResList := FExeImage.FCursorResources.List;
    for I := 0 to Cnt - 1 do
    begin
      ResOrd := ResData.wNameOrdinal;
      for J := 0 to ResList.Count - 1 do
      begin
        if ResOrd = ResList[J].FDirEntry.Name then
        begin
          CursorResource := ResList[J] as TCursorResEntry;
          CursorResource.FResInfo := ResData;
          FList.Add(CursorResource);
        end;
      end;
      Inc(ResData);
    end;
  end;
  Result := FList;
end;

end.
