
{*******************************************************}
{                                                       }
{       Borland Delphi Run-time Library                 }
{       WDOSX Interrupt Interface Unit                  }
{                                                       }
{       for Delphi 4, 5 WDosX DOS-Extender              }
{       Copyright (c) 2000 by Immo Wache                }
{       e-mail: immo.wache@t-online.de                  }
{                                                       }
{*******************************************************}

{
  Version history:
    IWA 11/20/00 Version 1.1
}

{ TODO : use Dpmi functions from unit WDosDpmi }

unit WDosInterrupts;

interface

uses
  WDosSysUtils, WDosPorts, WDosDpmi;

type
  DpmiPmVector = record
    PmOffset: Pointer;
    PmSelector: Integer;
  end;

procedure GetIntVec(IntNo: Integer; var Vector: DpmiPmVector); pascal;
procedure SetIntVec(IntNo: Integer; Vector: DpmiPmVector); pascal;
procedure SetIntProc(IntNo: Integer; Proc: TProcedure); pascal;
procedure CallInterrupt(Vector: DpmiPmVector); register; assembler;
procedure Reboot(Cold: Boolean);

implementation

const
  DPMIINT = $31;
  PostTest =$472;

var
  IntStack: array [0..$FFFF] of Byte;
  IntStackTop: LongInt;
  DataSelector: LongInt;
  Pic1Base, Pic2Base: Byte;
  IrqHandler: array [0..16] of TProcedure;
  ProcLookup: array [0..16] of Pointer;

procedure CallInterrupt(Vector: DpmiPmVector); register; assembler;
asm
	PUSHFD
	PUSH	CS
        MOV     EDX,DpmiPmVector [EAX].PmOffset
        MOV     DWORD PTR [OFFSET @@Ofs],EDX
        MOV     EDX,DpmiPmVector [EAX].PmSelector
        MOV     WORD PTR [OFFSET @@Sel],DX
        DB      $9A
  @@Ofs:
        DD      $0
  @@Sel:
        DW      $0
end;



procedure EnterInterrupt; register; assembler;
asm
	XCHG	ECX,[ESP]
	PUSH	EAX
	PUSH	EDX
	PUSH	GS
	PUSH	FS
	PUSH	ES
	PUSH	DS
	MOV	DS,DWORD PTR CS:[OFFSET DataSelector]
	MOV	EDX,SS
	MOV	EAX,ESP
	SUB	DWORD PTR [OFFSET IntStackTop],4000H
	PUSH	DS
	POP	SS
	MOV	ESP,DWORD PTR [OFFSET IntStackTop]
	PUSH	EDX
	MOV	EDX,DS
	MOV	FS,EDX
	MOV	ES,EDX
	PUSH	EAX
	CALL	ECX
	DB	0FH,0B2H,24H,24H // LSS ESP,[ESP]
	ADD	DWORD PTR [OFFSET IntStackTop],4000H
	POP	DS
	POP	ES
	POP	FS
	POP	GS
	POP	EDX
	POP	EAX
	POP	ECX
	IRETD
end;

procedure HandleIrq0; register; assembler;
asm
	CALL	EnterInterrupt
	CALL	DWORD PTR [OFFSET IrqHandler]
end;

procedure HandleIrq1; register; assembler;
asm
	CALL	EnterInterrupt
	CALL	DWORD PTR [OFFSET IrqHandler +4 *01]
end;

procedure HandleIrq2; register; assembler;
asm
	CALL	EnterInterrupt
	CALL	DWORD PTR [OFFSET IrqHandler +4 *02]
end;

procedure HandleIrq3; register; assembler;
asm
	CALL	EnterInterrupt
	CALL	DWORD PTR [OFFSET IrqHandler +4 *03]
end;

procedure HandleIrq4; register; assembler;
asm
	CALL	EnterInterrupt
	CALL	DWORD PTR [OFFSET IrqHandler +4 *04]
end;

procedure HandleIrq5; register; assembler;
asm
	CALL	EnterInterrupt
	CALL	DWORD PTR [OFFSET IrqHandler +4 *05]
end;

procedure HandleIrq6; register; assembler;
asm
	CALL	EnterInterrupt
	CALL	DWORD PTR [OFFSET IrqHandler +4 *06]
end;

procedure HandleIrq7; register; assembler;
asm
	CALL	EnterInterrupt
	CALL	DWORD PTR [OFFSET IrqHandler +4 *07]
end;

procedure HandleIrq8; register; assembler;
asm
	CALL	EnterInterrupt
	CALL	DWORD PTR [OFFSET IrqHandler +4 *08]
end;

procedure HandleIrq9; register; assembler;
asm
	CALL	EnterInterrupt
	CALL	DWORD PTR [OFFSET IrqHandler +4 *09]
end;

procedure HandleIrq10; register; assembler;
asm
	CALL	EnterInterrupt
	CALL	DWORD PTR [OFFSET IrqHandler +4 *10]
end;

procedure HandleIrq11; register; assembler;
asm
	CALL	EnterInterrupt
	CALL	DWORD PTR [OFFSET IrqHandler +4 *11]
end;

procedure HandleIrq12; register; assembler;
asm
	CALL	EnterInterrupt
	CALL	DWORD PTR [OFFSET IrqHandler +4 *12]
end;

procedure HandleIrq13; register; assembler;
asm
	CALL	EnterInterrupt
	CALL	DWORD PTR [OFFSET IrqHandler +4 *13]
end;

procedure HandleIrq14; register; assembler;
asm
	CALL	EnterInterrupt
	CALL	DWORD PTR [OFFSET IrqHandler +4 *14]
end;

procedure HandleIrq15; register; assembler;
asm
	CALL	EnterInterrupt
	CALL	DWORD PTR [OFFSET IrqHandler +4 *15]
end;

procedure HandleIrq16; register; assembler;
asm
	CALL	EnterInterrupt
	CALL	DWORD PTR [OFFSET IrqHandler +4 *16]
end;

function IsIrq(IntNr: Integer): Boolean; register;
begin
  Result:=(IntNr <$10) and (IntNr >$07) or (IntNr <$78) and (IntNr >$6F)
      or (IntNr =$1C);
end;

function IntToIrq(IntNr: Integer): Integer; register;
begin
  if IntNr =$1C then
    Result :=16
  else if IntNr >16 then
    Result :=IntNr -$68
  else Result :=IntNr -8;
end;

function TrueIntNr(IntNr: Integer): Integer; register;
begin
  if IsIrq(IntNr) then
  begin
    if IntNr =$1C then
      Result :=IntNr
    else if IntToIrq(IntNr) <8 then
      Result :=IntToIrq(IntNr) +Pic1Base
    else
      Result :=IntToIrq(IntNr) +Pic2Base -$08;
  end
  else Result :=IntNr;
end;

procedure InitInterrupts; assembler;
asm
	MOV	EAX, OFFSET IntStack
	ADD	EAX,10000H
	MOV	IntStackTop,EAX

	MOV	DWORD PTR [OFFSET ProcLookup +4 *00],OFFSET HandleIrq0
	MOV	DWORD PTR [OFFSET ProcLookup +4 *01],OFFSET HandleIrq1
	MOV	DWORD PTR [OFFSET ProcLookup +4 *02],OFFSET HandleIrq2
	MOV	DWORD PTR [OFFSET ProcLookup +4 *03],OFFSET HandleIrq3
	MOV	DWORD PTR [OFFSET ProcLookup +4 *04],OFFSET HandleIrq4
	MOV	DWORD PTR [OFFSET ProcLookup +4 *05],OFFSET HandleIrq5
	MOV	DWORD PTR [OFFSET ProcLookup +4 *06],OFFSET HandleIrq6
	MOV	DWORD PTR [OFFSET ProcLookup +4 *07],OFFSET HandleIrq7
	MOV	DWORD PTR [OFFSET ProcLookup +4 *08],OFFSET HandleIrq8
	MOV	DWORD PTR [OFFSET ProcLookup +4 *09],OFFSET HandleIrq9
	MOV	DWORD PTR [OFFSET ProcLookup +4 *10],OFFSET HandleIrq10
	MOV	DWORD PTR [OFFSET ProcLookup +4 *11],OFFSET HandleIrq11
	MOV	DWORD PTR [OFFSET ProcLookup +4 *12],OFFSET HandleIrq12
	MOV	DWORD PTR [OFFSET ProcLookup +4 *13],OFFSET HandleIrq13
	MOV	DWORD PTR [OFFSET ProcLookup +4 *14],OFFSET HandleIrq14
	MOV	DWORD PTR [OFFSET ProcLookup +4 *15],OFFSET HandleIrq15
	MOV	DWORD PTR [OFFSET ProcLookup +4 *16],OFFSET HandleIrq16

        PUSH	EBX
        PUSH	ESI
        MOV	EAX,400H
        INT	31H
        MOV	Pic1Base,DH
        MOV	Pic2Base,DL
        POP	ESI
        POP	EBX
        MOV	DataSelector,DS
end;

procedure GetIntVec(IntNo: Integer; var Vector: DpmiPmVector); pascal;
begin
  IntNo :=TrueIntNr(IntNo);
  asm
	PUSH	EBX
	MOV	EBX,IntNo
	MOV	EAX,204H
	INT	DPMIINT
        MOV     EAX,[Vector]
	MOV	DpmiPmVector [EAX].PmOffset,EDX
        mov     DpmiPmVector [EAX].PmSelector,ECX
	POP	EBX
  end;
end;

procedure SetIntVec(IntNo: Integer; Vector: DpmiPmVector); pascal;
begin
  IntNo :=TrueIntNr(IntNo);
  asm
	PUSH	EBX
	MOV	EBX,IntNo
	MOV	EDX,DpmiPmVector [Vector].PmOffset
	MOV	ECX,DpmiPmVector [Vector].PmSelector
	MOV	EAX,0205H
	INT	DPMIINT
	POP	EBX
  end;
end;

procedure SetIntProc(IntNo: Integer; Proc: TProcedure); pascal;
begin
  IntNo:=TrueIntNr(IntNo);
  if IsIrq(IntNo) then
  begin
    IrqHandler[IntToIrq(IntNo)] :=Proc;
    Proc :=ProcLookup[IntToIrq(IntNo)];
  end;
  asm
	PUSH	EBX
	MOV	EBX,IntNo
	MOV	EDX,Proc
	MOV	ECX,CS
	MOV	EAX,0205H
	INT	DPMIINT
	POP	EBX
  end;
end;

procedure Reboot(Cold: Boolean);
var
  Regs: DpmiRmRegisters;
begin
  if Cold then
    MemW[PostTest] :=0
  else
    MemW[PostTest] :=$1234;
  Fillchar(Regs, SizeOf(Regs), 0);
  with Regs do
  begin
    CS :=$FFFF;
    IP :=$0000;
  end;
  { enable interrupts }
  Port[$A0]:=$20;
  Port[$20]:=$20;
  DpmiRealModeIntCall(True, Regs, 0);
end;


initialization
  InitInterrupts;

end.
