
{*******************************************************}
{                                                       }
{       Borland Delphi WDosX Component Library          }
{       LPT Control Unit                                }
{                                                       }
{       for Delphi 4, 5 WDosX DOS-Extender              }
{       Copyright (c) 2000 by Immo Wache                }
{       e-mail: immo.wache@t-online.de                  }
{                                                       }
{*******************************************************}

{
  Version history:
    IWA 11/20/00 Version 1.1
}

unit WDosLptCtrls;

interface

{$IFDEF DESIGNPACKAGE}
uses
  SysUtils, Classes, Forms, WDosResStrings;
{$ELSE}
uses
  WDosSysUtils, WDosClasses, WDosForms, WDosResStrings;
{$ENDIF}

type
  TLptDevice =(ldNull, ldLPT1, ldLPT2, ldLPT3);

  TwdxLptCtrl = class (TComponent)
  private
    FActive: Boolean;
    FCtrl: Byte;
    FData: Byte;
    FDevice: TLptDevice;
    FDeviceAddr: Word;
    FStreamedActive: Boolean;
    function GetCtrl: Byte;
    function GetData: Byte;
    function GetStatus: Byte;
    procedure SetActive(Value: Boolean);
    procedure SetCtrl(Value: Byte);
    procedure SetData(Value: Byte);
    procedure SetDevice(Value: TLptDevice);
  protected
    procedure Loaded; override;
  public
    destructor Destroy; override;
    procedure Close;
    procedure Open;
    property Ctrl: Byte read GetCtrl write SetCtrl;
    property Data: Byte read GetData write SetData;
    property Status: Byte read GetStatus;
  published
    property Active: Boolean read FActive write SetActive;
    property Device: TLptDevice read FDevice write SetDevice;
  end;
  
  ELptCtrlError = class (Exception)
  end;
  
implementation

{$IFNDEF DESIGNPACKAGE}
uses
  WDosPorts;
{$ENDIF}

const
  LptDeviceNames: array[TLptDevice] of string =
      ('NULL:', 'LPT1:', 'LPT2:', 'LPT3:');

type
  TLptOpenDevice =ldLPT1..ldLPT3;
  TLptOpenDevices = set of TLptOpenDevice;

var
  LptOpenDevices: TLptOpenDevices;

{
************************************** TwdxLptCtrl ***************************************
}
destructor TwdxLptCtrl.Destroy;
begin
  { close device }
  Close;
  inherited Destroy;
end;

procedure TwdxLptCtrl.Close;
begin
  if FActive then
  begin
    Exclude(LptOpenDevices, FDevice);
    FActive :=False;
  end;
end;

procedure TwdxLptCtrl.Loaded;
begin
  inherited Loaded;
  try
    Active :=FStreamedActive;
  except
    { let Delphi handle exceptions in design modus }
    if csDesigning in ComponentState then
      Application.HandleException(Self)
    else
      raise;
  end;
end;

procedure TwdxLptCtrl.Open;
begin
  if not FActive then
  begin
    { check for a valid device }
    if (FDevice =ldNull) or (FDevice in LptOpenDevices) then
      raise ELptCtrlError.CreateFmt(sLptCtrlOpenError,
          [LptDeviceNames[FDevice]]);
    Include(LptOpenDevices, FDevice);
    {$IFNDEF DESIGNPACKAGE}
    { syncronize Ctrl and Data }
    FCtrl :=Port[FDeviceAddr +2];
    Port[FDeviceAddr] :=FData;
    {$ENDIF}
    { set Active }
    FActive :=True;
  end;
end;

procedure TwdxLptCtrl.SetActive(Value: Boolean);
begin
  if csLoading in ComponentState then
    FStreamedActive :=Value
  else
    if Value then Open else Close;
end;

function TwdxLptCtrl.GetCtrl: Byte;
begin
  if FActive then
    {$IFDEF DESIGNPACKAGE}
    Result :=0
    {$ELSE}
    Result :=Port[FDeviceAddr +2]
    {$ENDIF}
  else
    raise ELptCtrlError.CreateFmt(sLptCtrlNotOpen, [LptDeviceNames[FDevice]]);
end;

procedure TwdxLptCtrl.SetCtrl(Value: Byte);
begin
  if FActive then
  begin
    if FCtrl <>Value then
    begin
      {$IFNDEF DESIGNPACKAGE}
      Port[FDeviceAddr +2] :=Value;
      {$ENDIF}
      FCtrl :=Value;
    end;
  end
  else
    raise ELptCtrlError.CreateFmt(sLptCtrlNotOpen, [LptDeviceNames[FDevice]]);
end;

function TwdxLptCtrl.GetData: Byte;
begin
  if FActive then
    Result :=FData
  else
    raise ELptCtrlError.CreateFmt(sLptCtrlNotOpen, [LptDeviceNames[FDevice]]);
end;

procedure TwdxLptCtrl.SetData(Value: Byte);
begin
  if FActive then
  begin
    if FData <>Value then
    begin
      {$IFNDEF DESIGNPACKAGE}
      Port[FDeviceAddr] :=Value;
      {$ENDIF}
      FData :=Value;
    end;
  end
  else
    raise ELptCtrlError.CreateFmt(sLptCtrlNotOpen, [LptDeviceNames[FDevice]]);
end;

procedure TwdxLptCtrl.SetDevice(Value: TLptDevice);
  
  const
    Offset: array[TLptDevice] of Byte =(0, 8, 10, 12);
  
begin
  if Value <>FDevice then
  begin
    if not FActive then
    begin
      FDevice :=Value;
      { get port address of device }
      {$IFDEF DESIGNPACKAGE}
      FDeviceAddr :=0;
      {$ELSE}
      FDeviceAddr :=MemW[$400 +Offset[FDevice]];
      {$ENDIF}
    end
    else
      raise ELptCtrlError.CreateFmt(sLptCtrlOnOpenError,
          [LptDeviceNames[FDevice]]);
  end;
end;

function TwdxLptCtrl.GetStatus: Byte;
begin
  if FActive then
    {$IFDEF DESIGNPACKAGE}
    Result :=0
    {$ELSE}
    Result :=Port[FDeviceAddr +1]
    {$ENDIF}
  else
    raise ELptCtrlError.CreateFmt(sLptCtrlNotOpen, [LptDeviceNames[FDevice]]);
end;

end.
