unit WDosLptInOuts;

interface

uses
  SysUtils, Classes, WDosPlcs, WDosPlcUtils, WDosLptCtrls;

type
  TwdxLptInOut = class (TCustomInOut)
  private
    FLptCtrl: TwdxLptCtrl;
    function GetDevice: TLptDevice;
    procedure SetDevice(Value: TLptDevice);
  protected
    procedure BeforeDestroy; override;
    procedure DoClose; override;
    procedure DoOpen; override;
    procedure ProcessInput(var IoData: TIoData); override;
    procedure ProcessOutput(var IoData: TIoData); override;
    procedure ResetOutput; override;
  public
    constructor Create(aOwner: TComponent); override;
  published
    property Device: TLptDevice read GetDevice write SetDevice;
  end;
  
implementation

{
************************************** TwdxLptInOut **************************************
}
constructor TwdxLptInOut.Create(aOwner: TComponent);
begin
  inherited Create(aOwner);
  { create own objects }
  FLptCtrl := TwdxLptCtrl.Create(Self);
end;

procedure TwdxLptInOut.BeforeDestroy;
begin
  { free own objects }
  FLptCtrl.Free;
end;

procedure TwdxLptInOut.DoClose;
begin
  { close LPT device }
  FLptCtrl.Close;
end;

procedure TwdxLptInOut.DoOpen;
begin
  { open LPT device }
  FLptCtrl.Open;
end;

procedure TwdxLptInOut.ProcessInput(var IoData: TIoData);
begin
  { read inputs }
  Byte(IoData[0, False]) :=FLptCtrl.Status shr 3 xor $0F;
end;

procedure TwdxLptInOut.ProcessOutput(var IoData: TIoData);
begin
  { function works only in run state }
  if State =iosRun then
  begin
    { update outputs }
    FLptCtrl.Data :=Byte(IoData[0, True]);
    FLptCtrl.Ctrl :=Byte(IoData[1, True]) and $0F xor $0B;
  end
  else ResetOutput;
end;

procedure TwdxLptInOut.ResetOutput;
begin
  { reset outputs }
  FLptCtrl.Data :=0;
  FLptCtrl.Ctrl :=0 xor $0B;
end;

function TwdxLptInOut.GetDevice: TLptDevice;
begin
  Result := FLptCtrl.Device;
end;

procedure TwdxLptInOut.SetDevice(Value: TLptDevice);
begin
  FLptCtrl.Device := Value;
end;

end.
