
{*******************************************************}
{                                                       }
{       Borland Delphi Run-time Library                 }
{       Win32 Resource API Interface Unit               }
{                                                       }
{       for Delphi 4, 5 WDosX DOS-Extender              }
{       Copyright (c) 2000 by Immo Wache                }
{       e-mail: immo.wache@t-online.de                  }
{                                                       }
{*******************************************************}

{
  Version history:
    IWA 11/20/00 Version 1.1
}

unit WDosResources;

interface

uses
  WDosWindows, WDosSysUtils, WDosClasses, WDosWfseStreams, WDosExeImage,
  WDosRXTypes;

function FreeResource(hResData: HGLOBAL): BOOL; stdcall;
function LockResource(hResData: HGLOBAL): Pointer; stdcall;
function UnlockResource(hResData: THandle): BOOL;
function LoadResource(hModule: HINST; hResInfo: HRSRC): HGLOBAL; stdcall;
function SizeofResource(hModule: HINST; hResInfo: HRSRC): DWORD; stdcall;
function FindResourceA(hModule: HMODULE; lpName, lpType: PAnsiChar): HRSRC; stdcall;
function FindResource(hModule: HMODULE; lpName, lpType: PChar): HRSRC; stdcall;

implementation

const
  HandleOffset = 1024;
  GlobalOffset = 2048;

type
  TResourceWrapper = class (TObject)
  private
    FExeImage: TExeImage;
    FResourceItems: TList;
    function GetExeImage: TExeImage;
  protected
    property ExeImage: TExeImage read GetExeImage;
  public
    constructor Create;
    destructor Destroy; override;
    function FindResource(hModule: HMODULE; lpName, lpType: PChar): HRSRC;
    function FreeResource(hResData: HGLOBAL): BOOL;
    function LoadResource(hModule: HINST; hResInfo: HRSRC): HGLOBAL;
    function LockResource(hResData: HGLOBAL): Pointer;
    function SizeOfResource(hModule: HINST; hResInfo: HRSRC): DWORD;
    function UnlockResource(hResData: HGLOBAL): BOOL;
  end;

var
  FResourceWrapper: TResourceWrapper = nil;

function ResourceWrapper: TResourceWrapper;
begin
  if not Assigned(FResourceWrapper) then
    FResourceWrapper :=TResourceWrapper.Create;
  Result :=FResourceWrapper;
end;

function FindResourceA(hModule: HMODULE; lpName, lpType: PAnsiChar): HRSRC;
    stdcall;
begin
  Result :=ResourceWrapper.FindResource(hModule, lpName, lpType);
end;

function FindResource(hModule: HMODULE; lpName, lpType: PChar): HRSRC; stdcall;
begin
  Result :=FindResourceA(hModule, lpName, lpType);
end;

function SizeofResource(hModule: HINST; hResInfo: HRSRC): DWORD; stdcall;
begin
  Result :=ResourceWrapper.SizeOfResource(hModule, hResInfo);
end;

function LockResource(hResData: HGLOBAL): Pointer; stdcall;
begin
  Result :=ResourceWrapper.LockResource(hResData);
end;

function UnlockResource(hResData: THandle): BOOL;
begin
  Result :=ResourceWrapper.UnlockResource(hResData);
end;

function LoadResource(hModule: HINST; hResInfo: HRSRC): HGLOBAL; stdcall;
begin
  Result :=ResourceWrapper.LoadResource(hModule, hResInfo);
end;

function FreeResource(hResData: HGLOBAL): BOOL; stdcall;
begin
  Result :=ResourceWrapper.FreeResource(hResData);
end;

{
******************************* TResourceWrapper *******************************
}
constructor TResourceWrapper.Create;
begin
  inherited Create;
  FResourceItems :=TList.Create;
end;

destructor TResourceWrapper.Destroy;
begin
  FResourceItems.Free;
  FExeImage.Free;
  inherited Destroy;
end;

function TResourceWrapper.FindResource(hModule: HMODULE; lpName, lpType: 
    PChar): HRSRC;
var
  ResourceTypeItem: TResourceItem;
  ResourceItem: TResourceItem;
  Index: Integer;
begin
  Result :=0;
  if hModule <>HInstance then Exit;
  ResourceTypeItem :=ExeImage.GetResources(TResourceType(lpType));
  if not Assigned(ResourceTypeItem) then Exit;
  ResourceItem :=ResourceTypeItem.List.FindResourceItem(lpName);
  if not Assigned(ResourceItem) then Exit;
  with FResourceItems do
  begin
    Index :=IndexOf(ResourceItem);
    if Index =-1 then
      Index :=Add(ResourceItem);
  end;
  Result :=Index +HandleOffset;
end;

function TResourceWrapper.FreeResource(hResData: HGLOBAL): BOOL;
var
  Index: Integer;
  ResourceItem: TResourceItem;
begin
  Result :=False;
  Index :=hResData -GlobalOffset;
  with FResourceItems do
  begin
    if (Index <0) or (Index >=Count) then Exit;
    ResourceItem :=Items[Index];
    ResourceItem.FreeRawData;
    Delete(Index);
  end;
  Result :=True;
end;

function TResourceWrapper.LoadResource(hModule: HINST; hResInfo: HRSRC): 
    HGLOBAL;
var
  Index: Integer;
begin
  Result :=0;
  if hModule <>HInstance then Exit;
  Index :=hResInfo -HandleOffset;
  with FResourceItems do
    if (Index <0) or (Index >=Count) then Exit;
  Result :=Index +GlobalOffset;
end;

function TResourceWrapper.LockResource(hResData: HGLOBAL): Pointer;
var
  Index: Integer;
  ResourceItem: TResourceItem;
begin
  Result :=nil;
  Index :=hResData -GlobalOffset;
  with FResourceItems do
  begin
    if (Index <0) or (Index >=Count) then Exit;
    ResourceItem :=Items[Index];
  end;
  Result :=ResourceItem.RawData;
end;

function TResourceWrapper.SizeOfResource(hModule: HINST; hResInfo: HRSRC): 
    DWORD;
var
  Index: Integer;
  ResourceItem: TResourceItem;
begin
  Result :=0;
  if hModule <>HInstance then Exit;
  Index :=hResInfo -HandleOffset;
  with FResourceItems do
  begin
    if (Index <0) or (Index >=Count) then Exit;
    ResourceItem :=Items[Index];
  end;
  Result :=ResourceItem.Size;
end;

function TResourceWrapper.UnlockResource(hResData: HGLOBAL): BOOL;
var
  Index: Integer;
  ResourceItem: TResourceItem;
begin
  Result :=False;
  Index :=hResData -GlobalOffset;
  with FResourceItems do
  begin
    if (Index <0) or (Index >=Count) then Exit;
    ResourceItem :=Items[Index];
  end;
  ResourceItem.FreeRawData;
  Result :=True;
end;

function TResourceWrapper.GetExeImage: TExeImage;
begin
  if not Assigned(FExeImage) then
    FExeImage :=TExeImage.CreateImage(nil, TWfseStream.Create('WdosxMain'));
  Result :=FExeImage;
end;

initialization

finalization
  FResourceWrapper.Free;

end.
